// ImageView.cpp: implementation of the ImageView class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageView.h"
#include <math.h>
#include <float.h>

#ifdef USE_RENDER_THREAD
#include <process.h>
#include "RenderProc.cpp"
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CImageView::CImageView(CComQIPtr<IPanImage> img, bool tVarying, CPerfData* pData)
: m_image(img),m_perfData(pData),m_imageBuffer(NULL),m_width(0),m_height(0)
{
    m_imageBuffer = NULL;
    m_timeVarying = tVarying;

    if (m_timeVarying) {
        m_startMS = timeGetTime();
    }
    
    //
    // Query the Pan image for what viewing transform values
    // to use initially. Why? Because the component may have been
    // instantiated from a settings file, where the viewing transform
    // might be something other than the identity.
    //

    // The viewer uses doubles for the view transform, the component float.
    // ToDo: make them agree..?
    float fZoom, fPanX, fPanY;
    m_image->GetViewTransform(&fPanX,&fPanY, &fZoom);
    m_panX = (double)fPanX;
    m_panY = (double)fPanY;
    m_zoom = (double)fZoom;

#ifdef USE_RENDER_THREAD
  UINT threadID;
  ThreadData* tData = new ThreadData();

  tData->hStart    = m_hRenderStart    = CreateEvent(NULL, FALSE, FALSE, NULL);
  tData->hComplete = m_hRenderComplete = CreateEvent(NULL, FALSE, FALSE, NULL);
  tData->hQuit     = m_hRenderQuit     = CreateEvent(NULL, FALSE, FALSE, NULL);
  m_rData = &(tData->rData);
  
  // Note: this will only work if you're compiling with /MT(d) to get at
  // the 
  m_hThread = (HANDLE)_beginthreadex(NULL,0,RenderProc, tData, 0, &threadID);
  // For now, assume thread creation didn't run into any problems.
#endif

  if (m_perfData) {
      m_perfData->EnableFPS(m_timeVarying);
  }
}

CImageView::~CImageView()
{
    if (m_imageBuffer)
       free(m_imageBuffer);

#ifdef USE_RENDER_THREAD
   // signal the rendering thread to shutdown.
   SetEvent(m_hRenderQuit);
   CloseHandle(m_hRenderStart);
   CloseHandle(m_hRenderComplete);
   CloseHandle(m_hRenderQuit);
   CloseHandle(m_hThread);
#endif
}


// Random float in [-1,1].  ToDo: replace with version that gives a
// Gaussian distribution, for better anti-aliasing.
static float frand()
{
    const int randTop = RAND_MAX/2-1;
    return (float)((rand() % (2*randTop+1)) - randTop) / randTop;
}
    
// Just BLiT the computed pixels onto the view.
void
CImageView::Blit(HWND hWnd, HDC hdc, RECT& rc, int width, int height)
{
    // This seems like a lot of work just to blit some already-computed
    // pixels.  Also, we're blitting all of the pixels, but shouldn't we
    // do just those in rc?  Well, SetDIBits doesn't understand rectangles
    // of pixels, but we could at least do just the needed number of scan
    // lines.
    HDC hdcMem;
    HBITMAP hBitmap, hOldBmp;

    BITMAPINFO bmi;
    BITMAPINFOHEADER bmih;

    bmih.biSize          = sizeof(BITMAPINFOHEADER);
    bmih.biWidth         = width;
    bmih.biHeight        = height;   // or negate for "top-down" DIB
    bmih.biPlanes        = 1;
    bmih.biBitCount      = 32;
    bmih.biCompression   = BI_RGB;
    bmih.biSizeImage     = 0;
    bmih.biXPelsPerMeter = 0;
    bmih.biYPelsPerMeter = 0;
    bmih.biClrUsed       = 0;
    bmih.biClrImportant  = 0;

    bmi.bmiHeader = bmih;

    hdcMem = CreateCompatibleDC(hdc);
    hBitmap = CreateCompatibleBitmap(hdc, width, height);

    SetDIBits(hdcMem, hBitmap, 0, height, m_imageBuffer, &bmi, 0);

    // BLT sampled image into the viewer window.
    hOldBmp = (HBITMAP)SelectObject(hdcMem, hBitmap);

    int rcWidth  = rc.right  - rc.left,
        rcHeight = rc.bottom - rc.top;
    BitBlt(hdc, rc.left, rc.top, rcWidth, rcHeight,
           hdcMem, rc.left, rc.top, SRCCOPY);

    if (hOldBmp)  SelectObject(hdcMem, hOldBmp);
    DeleteObject(hBitmap);
    DeleteDC(hdcMem);
}



// Render the image (using the current viewing transform). 
void
CImageView::Render(RECT& rc, int width, int height,
                   int pass, double maxJitter)
{
    LARGE_INTEGER tStart, tEnd, freq;
    unsigned int cycles;
    int rcWidth  = rc.right  - rc.left,
        rcHeight = rc.bottom - rc.top;

    // By default, the range of the image function is an RGBQUAD, i.e., a
    // 32-bit value. Should you want to change that and use DDBs instead
    // (see below), you need to set 'bytesPixel' to fit the display mode
    // you're using., e.g.,
    //
    // int bitsPixel  = Win32.GetDeviceCaps(hdcMem, BITSPIXEL);
    // int bytesPixel = bitsPixel / 8;
    // 
    int bytesPixel = 4;

    // Compute sampling time.  Ignored if not timeVarying
    double t;
    
    // To really squeeze out a couple of extra cycles, we could
    // separate time-varying image into a sep. class, so that
    // we didn't have to perform the below test on a per-frame basis.
    if (m_timeVarying) {
        t = (timeGetTime() - m_startMS)/1000.0;
    } else {
        t = 0.0;
    }
  
    // The Pan display function renders into a back buffer of RGBQUADs, so
    // if there's been a change in view size, we better reallocate the
    // back buffer. (How badly are we fragmenting the heap?)
    if (m_imageBuffer != 0 && (width != m_width || height != m_height))
        free(m_imageBuffer);

    if (width != m_width || height != m_height)
        m_imageBuffer = (RGBQUAD*)malloc(height * width * bytesPixel);

    m_width = width;
    m_height = height;

    // for performance timing.
    if (m_perfData) {
        QueryPerformanceCounter(&tStart);
    }

 
    // We assume that the generated code will have been compiled with
    // /QIfist, i.e., it won't be calling _ftol() to convert floating
    // point constants to ints. Setting the rounding mode to trunc here
    // may badly affect the results of other fp computations that the
    // displayer does, so we may have to back out of using /QIfist.
    // Round towards minus infinity:
    unsigned int oMode = _controlfp ( _MCW_RC, _RC_DOWN );

    m_image->SetViewSize(width,height);

    // Constant for now.  ToDo: compute sensibly.
    const float displayWeight = (float) 1.0 / pass;
    // Max jitter in pixels, but not in the first pass, so as not to get
    // wobbling when panning, zooming, and adjusting sliders.
    const double maxJitterPixels = pass == 1 ? (double) 0 : maxJitter;
    // Max jitter in image space
    double maxJitterIm = (double) maxJitterPixels * m_zoom;

    // Add some sub-pixel jitter for anti-aliasing.
    double was_panX = m_panX, was_panY = m_panY;
    double dx = frand()*maxJitterIm, dy = frand()*maxJitterIm;
    SetTranslation(m_panX+dx,m_panY+dy);

#ifdef USE_RENDER_THREAD
    RenderData* rData = (RenderData*)m_rData;

    rData->t       = (float)t;
    rData->xStart  = rc.left;
    rData->yStart  = height - rc.bottom;
    rData->rWidth  = rcWidth;
    rData->rHeight = rcHeight/2;
    rData->weight  = displayWeight;
    rData->imageBuffer = (int)m_imageBuffer;
    rData->disper = m_image;

    // Signal that the rendering thread should start.
    SetEvent(m_hRenderStart);

    m_image->Display
        ( (float)t,
          rc.left,  height - rc.bottom + rcHeight/2, 
          rcWidth, rcHeight/2,
          displayWeight, 0, (int)m_imageBuffer
            );
		   
    // reset the rounding mode.
    _controlfp(  _MCW_RC, (oMode &  _MCW_RC));

    // Wait for rendering thread to finish its portion..
    WaitForSingleObject(m_hRenderComplete, INFINITE);
#else
    m_image->Display
        ( (float)t,
          rc.left,  height - rc.bottom,
          rcWidth, rcHeight,
          displayWeight, 0, (int)m_imageBuffer
            );
#endif

    // gather performance data to compute the 'cost' of rendering
    // one frame.
    if (m_perfData) {
        QueryPerformanceCounter(&tEnd);
        cycles = (unsigned int)(tEnd.QuadPart - tStart.QuadPart);
      
        QueryPerformanceFrequency(&freq);
        int numPixels = rcWidth * rcHeight;
        float pps = (float)numPixels*((float)freq.QuadPart/(float)cycles);
        m_perfData->SetPixelsPerSec(pps, pass);
    }

    // Eliminate jitter.
    SetTranslation(was_panX,was_panY);
}


// Update generated image to reflect change in viewing transform, time
// (if time varying.)
void
CImageView::Redisplay()
{
    if (m_perfData) {
      m_startMS = timeGetTime();
      m_perfData->EnableFPS(m_timeVarying);
    }
}
