module TransformE where

import Exp
import PointE
import Int

type TransformE = PointE -> PointE

rotateP :: FloatE -> TransformE
rotateP ang = \ (x,y) -> (x * c - y * s, y * c + x * s)
 where c = cos ang
       s = sin ang

translateP, scaleP :: VectorE -> TransformE
translateP (dx,dy) = \ (x,y) -> (x + dx, y + dy)
scaleP     (sx,sy) = \ (x,y) -> (sx * x, sy * y)

uscaleP :: FloatE -> TransformE
uscaleP s = scaleP (s,s)

shearP :: FloatE -> TransformE
shearP dxdy (x,y) = (x + dxdy*y, y)


-- Replicator with given tile size
tileP :: VectorE -> TransformE
tileP (w,h) = \ (x,y) -> (wrap' w x, wrap' h y)

-- Perform a transform on the polar form
polarXf :: TransformE -> TransformE
polarXf xf = fromPolar . xf . toPolar

-- Swirling transform.  Rotate each point by an amount proportional to
-- its distance from the origin.  Distance of r yields full rotation.
swirlP :: FloatE -> TransformE
swirlP r = \ p -> rotateP (distO p * (2 * pi / r)) p

-- The following form is a bit simpler (assuming polorXf), though
-- probably less efficient.
--swirlP r = polarXf (\ (rho,theta) -> (rho, theta + rho * (2 * pi / r)))

-- Ripple along the radius.  For f, try sin, tan, and whatever you think
-- of.  Designed for period of 2*pi.
cwaveP :: (FloatE -> FloatE) -> FloatE -> TransformE
cwaveP f r p = uscaleP (f (distO p * (2 * pi / r))) p

-- Wrap x to range [0,w).  Alternative wraps to [-w/2,w/2).
wrap, wrap' :: FloatE -> FloatE -> FloatE
wrap  w = \ x -> w * fracPart (x/w)
wrap' w = \ x -> wrap w (x + w2) - w2
  where w2 = w / 2


-- Transform "about" a point.
xfAbout :: PointE -> TransformE -> TransformE
xfAbout (x,y) xf = translateP (x,y) . xf . translateP (-x,-y)

-- 1D transform
type Transform1E = FloatE -> FloatE

radialXf  :: Transform1E -> TransformE
angularXf :: Transform1E -> TransformE

radialXf  f = polarXf (\ (r,a) -> (f r, a))
angularXf f = polarXf (\ (r,a) -> (r, f a))

radialScaleXf :: Transform1E -> TransformE
radialScaleXf f = radialXf $ \ r -> r * f r



{-
-- These aren't being used

rotateAbout :: PointE -> DoubleE -> TransformE
rotateAbout p ang = xfAbout p (rotate ang)

scaleAbout :: PointE -> (DoubleE,DoubleE) -> TransformE
scaleAbout p sc = xfAbout p (scaleP sc)

uscaleAbout :: PointE -> DoubleE -> TransformE
uscaleAbout p sc = scaleAbout p (sc,sc)
-}

-- Radially rippling transform.  Arguments: number of lobes; radius perturbation max.
rippleRadiusP' :: IntE -> FloatE -> TransformE
rippleRadiusP' n s = polarXf xf
 where
   xf (rho,theta) = (rho + s * sin (i2f n * theta), theta)

-- The next def gives proportional rippling, which often looks better.
rippleRadiusP :: IntE -> FloatE -> TransformE
rippleRadiusP n s = polarXf xf
 where
   xf (rho,theta) = (rho * (1 + s * sin (i2f n * theta)), theta)

-- Takes radius of effect and number of full revolutions at center
-- (positive for cw, negative for ccw).  Swirls inside or outside part of
-- circle, depending on first argument.
iswirling :: BoolE -> FloatE -> FloatE -> TransformE
iswirling inside rad sw p = rotateP a p
 where
   d = rad - distO p 
   a = ifE (mbNot (d >=* 0)) 0 (2 * pi * sw * d / rad)
   mbNot = ifE inside notE id

-- Radial inversion
radInvertP :: TransformE
radInvertP = polarXf (\ (r,th) -> (1/r,th))
